@@ -255,6 +255,10 @@ class MuminDataset:
 
         logger.info('Loading dataset')
 
+        # Reset `nodes` and `relations` to ensure a fresh start
+        self.nodes = dict()
+        self.rels = dict()
+
         # Loop over the files in the dataset directory
         csv_paths = [path for path in self.dataset_dir.iterdir()
                      if str(path)[-4:] == '.csv']
@@ -298,14 +302,16 @@ class MuminDataset:
 
     def _shrink_dataset(self):
         '''Shrink dataset if `size` is 'small' or 'medium'''
-        if self.size == 'small' or self.size == 'medium':
+        if self.size != 'large':
             logger.info('Shrinking dataset')
 
             # Define the `relevance` threshold
             if self.size == 'small':
                 threshold = 0.80
-            else:
+            elif self.size == 'medium':
                 threshold = 0.75
+            elif self.size == 'test':
+                threshold = 0.99
 
             # Filter (:Tweet)-[:DISCUSSES]->(:Claim)
             discusses_rel = (self.rels[('tweet', 'discusses', 'claim')]
@@ -316,7 +322,26 @@ class MuminDataset:
             # Filter tweets
             tweet_df = self.nodes['tweet']
             include_tweet = tweet_df.tweet_id.isin(discusses_rel.src.tolist())
-            self.nodes['tweet'] = (tweet_df[include_tweet]
+            tweet_df = tweet_df[include_tweet].reset_index(drop=True)
+            self.nodes['tweet'] = tweet_df
+
+            # Filter (:Reply)-[:REPLY_TO]->(:Tweet)
+            reply_rel = self.rels[('reply', 'reply_to', 'tweet')]
+            include = reply_rel.tgt.isin(tweet_df.tweet_id.tolist())
+            reply_rel = reply_rel[include].reset_index(drop=True)
+            self.rels[('reply', 'reply_to', 'tweet')] = reply_rel
+
+            # Filter (:Reply)-[:QUOTE_OF]->(:Tweet)
+            quote_rel = self.rels[('reply', 'quote_of', 'tweet')]
+            include = quote_rel.tgt.isin(tweet_df.tweet_id.tolist())
+            quote_rel = quote_rel[include].reset_index(drop=True)
+            self.rels[('reply', 'quote_of', 'tweet')] = quote_rel
+
+            # Filter replies
+            reply_df = self.nodes['reply']
+            include_reply = reply_df.tweet_id.isin(reply_rel.src.tolist())
+            include_quote = reply_df.tweet_id.isin(quote_rel.src.tolist())
+            self.nodes['reply'] = (reply_df[include_reply | include_quote]
                                    .reset_index(drop=True))
 
             # Filter claims
@@ -345,6 +370,14 @@ class MuminDataset:
             posted_rel = posted_rel.reset_index(drop=True)
             self.rels[('user', 'posted', 'tweet')] = posted_rel
 
+            # Filter (:User)-[:POSTED]->(:Reply)
+            rposted_rel = self.rels[('user', 'posted', 'reply')]
+            rposted_rel = rposted_rel[rposted_rel.tgt.isin(self.nodes['reply']
+                                                               .tweet_id
+                                                               .tolist())]
+            rposted_rel = rposted_rel.reset_index(drop=True)
+            self.rels[('user', 'posted', 'reply')] = rposted_rel
+
             # Filter (:Tweet)-[:MENTIONS]->(:User)
             mentions_rel = self.rels[('tweet', 'mentions', 'user')]
             mentions_rel = mentions_rel[mentions_rel
@@ -355,11 +388,30 @@ class MuminDataset:
             mentions_rel = mentions_rel.reset_index(drop=True)
             self.rels[('tweet', 'mentions', 'user')] = mentions_rel
 
+            # Filter (:User)-[:FOLLOWS]->(:User)
+            follows_rel = self.rels[('user', 'follows', 'user')]
+            user_df = self.nodes['user']
+            has_posted = user_df.user_id.isin(posted_rel.src.tolist())
+            has_rposted = user_df.user_id.isin(rposted_rel.src.tolist())
+            was_mentioned = user_df.user_id.isin(mentions_rel.tgt.tolist())
+            filtered_users = (user_df[has_posted | has_rposted | was_mentioned]
+                              .reset_index(drop=True)
+                              .user_id
+                              .tolist())
+            follows_rel = follows_rel[follows_rel.src.isin(filtered_users)]
+            follows_rel = follows_rel.reset_index(drop=True)
+            self.rels[('user', 'follows', 'user')] = follows_rel
+
             # Filter users
             user_df = self.nodes['user']
             has_posted = user_df.user_id.isin(posted_rel.src.tolist())
+            has_rposted = user_df.user_id.isin(rposted_rel.src.tolist())
             was_mentioned = user_df.user_id.isin(mentions_rel.tgt.tolist())
-            self.nodes['user'] = (user_df[has_posted | was_mentioned]
+            is_followed = user_df.user_id.isin(follows_rel.tgt.tolist())
+            self.nodes['user'] = (user_df[has_posted |
+                                          has_rposted |
+                                          was_mentioned |
+                                          is_followed]
                                   .reset_index(drop=True))
 
             # Filter (:User)-[:MENTIONS]->(:User)
@@ -394,10 +446,15 @@ class MuminDataset:
             tweet_dfs = self._twitter.rehydrate_tweets(tweet_ids=tweet_ids)
 
             # Extract and store tweets and users
-            self.nodes[node_type] = tweet_dfs['tweets']
+            self.nodes[node_type] = (tweet_dfs['tweets']
+                                     .drop_duplicates(subset='tweet_id')
+                                     .reset_index(drop=True))
             if ('user' in self.nodes.keys() and
                     'username' in self.nodes['user'].columns):
-                user_df = self.nodes['user'].append(tweet_dfs['users'])
+                user_df = (self.nodes['user']
+                               .append(tweet_dfs['users'])
+                               .drop_duplicates(subset='user_id')
+                               .reset_index(drop=True))
             else:
                 user_df = tweet_dfs['users']
             self.nodes['user'] = user_df
@@ -412,16 +469,24 @@ class MuminDataset:
                             .rename(columns=dict(preview_image_url='url')))
                 image_df = (tweet_dfs['media']
                             .query('type == "photo"')
-                            .append(video_df))
+                            .append(video_df)
+                            .drop_duplicates(subset='media_key')
+                            .reset_index(drop=True))
 
                 if 'media' in self.nodes.keys():
-                    image_df = self.nodes['media'].append(image_df)
+                    image_df = (self.nodes['media']
+                                    .append(image_df)
+                                    .drop_duplicates(subset='media_key')
+                                    .reset_index(drop=True))
                 self.nodes['image'] = image_df
 
             # Extract and store polls
             if self.include_polls and len(tweet_dfs['polls']):
                 if 'poll' in self.nodes.keys():
-                    poll_df = self.nodes['poll'].append(tweet_dfs['polls'])
+                    poll_df = (self.nodes['poll']
+                                   .append(tweet_dfs['polls'])
+                                   .drop_duplicates(subset='poll_id')
+                                   .reset_index(drop=True))
                 else:
                     poll_df = tweet_dfs['polls']
                 self.nodes['poll'] = poll_df
@@ -429,7 +494,10 @@ class MuminDataset:
             # Extract and store places
             if self.include_places and len(tweet_dfs['places']):
                 if 'place' in self.nodes.keys():
-                    place_df = self.nodes['place'].append(tweet_dfs['places'])
+                    place_df = (self.nodes['place']
+                                    .append(tweet_dfs['places'])
+                                    .drop_duplicates(subset='place_id')
+                                    .reset_index(drop=True))
                 else:
                     place_df = tweet_dfs['places']
                 self.nodes['place'] = place_df
@@ -524,25 +592,6 @@ class MuminDataset:
             rel_df = pd.DataFrame(data_dict)
             self.rels[rel_type] = rel_df
 
-        # Update the (:Reply)-[:REPLY_TO]->(:Reply) relation
-        rel_type = ('reply', 'reply_to', 'reply')
-        if rel_type in self.rels.keys():
-            rel = self.rels[rel_type]
-            merged = (rel.merge(self.nodes['reply'][['tweet_id']]
-                                    .reset_index()
-                                    .rename(columns=dict(index='reply_idx1')),
-                                left_on='src',
-                                right_on='tweet_id')
-                         .merge(self.nodes['reply'][['tweet_id']]
-                                    .reset_index()
-                                    .rename(columns=dict(index='reply_idx2')),
-                                left_on='tgt',
-                                right_on='tweet_id'))
-            data_dict = dict(src=merged.reply_idx1.tolist(),
-                             tgt=merged.reply_idx2.tolist())
-            rel_df = pd.DataFrame(data_dict)
-            self.rels[rel_type] = rel_df
-
         # Update the (:Reply)-[:QUOTE_OF]->(:Tweet) relation
         rel_type = ('reply', 'quote_of', 'tweet')
         if rel_type in self.rels.keys():
@@ -562,6 +611,25 @@ class MuminDataset:
             rel_df = pd.DataFrame(data_dict)
             self.rels[rel_type] = rel_df
 
+        # Update the (:Reply)-[:REPLY_TO]->(:Reply) relation
+        rel_type = ('reply', 'reply_to', 'reply')
+        if rel_type in self.rels.keys():
+            rel = self.rels[rel_type]
+            merged = (rel.merge(self.nodes['reply'][['tweet_id']]
+                                    .reset_index()
+                                    .rename(columns=dict(index='reply_idx1')),
+                                left_on='src',
+                                right_on='tweet_id')
+                         .merge(self.nodes['reply'][['tweet_id']]
+                                    .reset_index()
+                                    .rename(columns=dict(index='reply_idx2')),
+                                left_on='tgt',
+                                right_on='tweet_id'))
+            data_dict = dict(src=merged.reply_idx1.tolist(),
+                             tgt=merged.reply_idx2.tolist())
+            rel_df = pd.DataFrame(data_dict)
+            self.rels[rel_type] = rel_df
+
         # Update the (:User)-[:RETWEETED]->(:Tweet) relation
         rel_type = ('user', 'retweeted', 'tweet')
         if rel_type in self.rels.keys():
@@ -626,9 +694,10 @@ class MuminDataset:
                                 .tolist())
                 node_df = pd.DataFrame(dict(tag=hashtags))
                 if 'hashtag' in self.nodes.keys():
-                    node_df = (self.nodes['hashtag'].append(node_df)
-                                                    .drop_duplicates()
-                                                    .reset_index(drop=True))
+                    node_df = (self.nodes['hashtag']
+                                   .append(node_df)
+                                   .drop_duplicates()
+                                   .reset_index(drop=True))
                 self.nodes['hashtag'] = node_df
 
             # Add hashtags from users
@@ -640,9 +709,10 @@ class MuminDataset:
                                 .tolist())
                 node_df = pd.DataFrame(dict(tag=hashtags))
                 if 'hashtag' in self.nodes.keys():
-                    node_df = (self.nodes['hashtag'].append(node_df)
-                                                    .drop_duplicates()
-                                                    .reset_index(drop=True))
+                    node_df = (self.nodes['hashtag']
+                                   .append(node_df)
+                                   .drop_duplicates()
+                                   .reset_index(drop=True))
                 self.nodes['hashtag'] = node_df
 
         # Add urls from tweets
@@ -657,9 +727,10 @@ class MuminDataset:
                         .tolist())
             node_df = pd.DataFrame(dict(url=urls))
             if 'url' in self.nodes.keys():
-                node_df = (self.nodes['url'].append(node_df)
-                                            .drop_duplicates()
-                                            .reset_index(drop=True))
+                node_df = (self.nodes['url']
+                               .append(node_df)
+                               .drop_duplicates()
+                               .reset_index(drop=True))
             self.nodes['url'] = node_df
 
         # Add urls from user urls
@@ -674,9 +745,10 @@ class MuminDataset:
                         .tolist())
             node_df = pd.DataFrame(dict(url=urls))
             if 'url' in self.nodes.keys():
-                node_df = (self.nodes['url'].append(node_df)
-                                            .drop_duplicates()
-                                            .reset_index(drop=True))
+                node_df = (self.nodes['url']
+                               .append(node_df)
+                               .drop_duplicates()
+                               .reset_index(drop=True))
             self.nodes['url'] = node_df
 
         # Add urls from user descriptions
@@ -691,9 +763,10 @@ class MuminDataset:
                         .tolist())
             node_df = pd.DataFrame(dict(url=urls))
             if 'url' in self.nodes.keys():
-                node_df = (self.nodes['url'].append(node_df)
-                                            .drop_duplicates()
-                                            .reset_index(drop=True))
+                node_df = (self.nodes['url']
+                               .append(node_df)
+                               .drop_duplicates()
+                               .reset_index(drop=True))
             self.nodes['url'] = node_df
 
         # Add urls from profile pictures
@@ -707,9 +780,10 @@ class MuminDataset:
                         .tolist())
             node_df = pd.DataFrame(dict(url=urls))
             if 'url' in self.nodes.keys():
-                node_df = (self.nodes['url'].append(node_df)
-                                            .drop_duplicates()
-                                            .reset_index(drop=True))
+                node_df = (self.nodes['url']
+                               .append(node_df)
+                               .drop_duplicates()
+                               .reset_index(drop=True))
             self.nodes['url'] = node_df
 
         # Add urls from articles
@@ -717,9 +791,10 @@ class MuminDataset:
             urls = self.nodes['article'].url.dropna().tolist()
             node_df = pd.DataFrame(dict(url=urls))
             if 'url' in self.nodes.keys():
-                node_df = (self.nodes['url'].append(node_df)
-                                            .drop_duplicates()
-                                            .reset_index(drop=True))
+                node_df = (self.nodes['url']
+                               .append(node_df)
+                               .drop_duplicates()
+                               .reset_index(drop=True))
             self.nodes['url'] = node_df
 
         # Add place features
@@ -983,7 +1058,9 @@ class MuminDataset:
                                      right_on='url'))
                 data_dict = dict(src=merged.user_idx.tolist(),
                                  tgt=merged.ul_idx.tolist())
-                rel_df = rel_df.append(pd.DataFrame(data_dict))
+                rel_df = (rel_df.append(pd.DataFrame(data_dict))
+                                .drop_duplicates()
+                                .reset_index(drop=True))
 
             if desc_urls_exist:
                 merged = (self.nodes['user'][['entities.description.urls']]
@@ -999,7 +1076,9 @@ class MuminDataset:
                                      right_on='url'))
                 data_dict = dict(src=merged.user_idx.tolist(),
                                  tgt=merged.ul_idx.tolist())
-                rel_df = rel_df.append(pd.DataFrame(data_dict))
+                rel_df = (rel_df.append(pd.DataFrame(data_dict))
+                                .drop_duplicates()
+                                .reset_index(drop=True))
 
             self.rels[('tweet', 'has_url', 'url')] = rel_df
 
@@ -1214,15 +1293,16 @@ class MuminDataset:
         return self
 
     def add_embeddings(self,
-                       nodes_to_embed: List[str] = ['tweet', 'user', 'claim',
-                                                    'article', 'image']):
+                       nodes_to_embed: List[str] = ['tweet', 'reply', 'user',
+                                                    'claim', 'article',
+                                                    'image']):
         '''Computes, stores and dumps embeddings of node features.
 
         Args:
             nodes_to_embed (list of str):
                 The node types which needs to be embedded. If a node type does
                 not exist in the graph it will be ignored. Defaults to
-                ['tweet', 'user', 'claim', 'article', 'image'].
+                ['tweet', 'reply', 'user', 'claim', 'article', 'image'].
         '''
         # Throw error if `transformers` has not been installed
         try:
@@ -1239,6 +1319,10 @@ class MuminDataset:
         if 'tweet' in nodes_to_embed:
             self._embed_tweets()
 
+        # Embed replies
+        if 'reply' in nodes_to_embed:
+            self._embed_replies()
+
         # Embed users
         if 'user' in nodes_to_embed:
             self._embed_users()
@@ -1289,6 +1373,35 @@ class MuminDataset:
 
         return self
 
+    def _embed_replies(self):
+        '''Embeds all the replies in the dataset'''
+        import transformers
+
+        logger.info('Embedding replies')
+
+        # Load text embedding model
+        model_id = self.text_embedding_model_id
+        pipeline = transformers.pipeline(task='feature-extraction',
+                                         model=model_id,
+                                         tokenizer=model_id)
+
+        # Define embedding function
+        def embed(text: str):
+            '''Extract a text embedding'''
+            return np.asarray(pipeline(text))[0, 0, :]
+
+        # Embed tweet text using the pretrained transformer
+        text_embs = self.nodes['reply'].text.progress_apply(embed)
+        self.nodes['reply']['text_emb'] = text_embs
+
+        # Embed tweet language using a one-hot encoding
+        languages = self.nodes['reply'].lang.tolist()
+        one_hotted = [np.asarray(lst)
+                      for lst in pd.get_dummies(languages).to_numpy().tolist()]
+        self.nodes['reply']['lang_emb'] = one_hotted
+
+        return self
+
     def _embed_users(self):
         '''Embeds all the users in the dataset'''
         import transformers
@@ -1302,13 +1415,18 @@ class MuminDataset:
                                          tokenizer=model_id)
 
         # Define embedding function
-        def embed(text: str):
+        def embed(text: Union[float, str]):
             '''Extract a text embedding'''
-            return np.asarray(pipeline(text))[0, 0, :]
+            if text != text:
+                return None
+            else:
+                return np.asarray(pipeline(text))[0, 0, :]
 
         # Embed user description using the pretrained transformer
         desc_embs = self.nodes['user'].description.progress_apply(embed)
-        self.nodes['tweet']['description_emb'] = desc_embs
+        emb_dim = desc_embs.dropna().iloc[0].shape
+        desc_embs[desc_embs.isna()] = np.zeros(emb_dim)
+        self.nodes['user']['description_emb'] = desc_embs
 
         return self
 
@@ -1424,6 +1542,10 @@ class MuminDataset:
                                  'source', 'public_metrics.retweet_count',
                                  'public_metrics.reply_count',
                                  'public_metrics.quote_count'],
+                          reply=['tweet_id', 'text', 'created_at', 'lang',
+                                 'source', 'public_metrics.retweet_count',
+                                 'public_metrics.reply_count',
+                                 'public_metrics.quote_count'],
                           user=['user_id', 'verified', 'protected',
                                 'created_at', 'username', 'description', 'url',
                                 'name', 'public_metrics.followers_count',
@@ -1582,14 +1704,14 @@ class MuminDataset:
                     node_df = (self.nodes[node_type]
                                    .rename(columns=dict(index='old_idx'))
                                    .reset_index())
-                    rel_df = (rel_df.merge(node_df,
+
+                    rel_df = (rel_df.merge(node_df[['index', 'old_idx']],
                                            left_on='src',
                                            right_on='old_idx')
                                     .drop(columns=['src', 'old_idx'])
                                     .rename(columns=dict(index='src')))
-                    self.rels[rel_type] = rel_df
-                    self.nodes[node_type] = (self.nodes[node_type]
-                                                 .drop(columns='index'))
+                    self.rels[rel_type] = rel_df[['src', 'tgt']]
+                    self.nodes[node_type] = self.nodes[node_type]
 
                 # If islands have been removed from the target, then update
                 # those indices
@@ -1597,14 +1719,17 @@ class MuminDataset:
                     node_df = (self.nodes[node_type]
                                    .rename(columns=dict(index='old_idx'))
                                    .reset_index())
-                    rel_df = (rel_df.merge(node_df,
+                    rel_df = (rel_df.merge(node_df[['index', 'old_idx']],
                                            left_on='tgt',
                                            right_on='old_idx')
                                     .drop(columns=['tgt', 'old_idx'])
                                     .rename(columns=dict(index='tgt')))
-                    self.rels[rel_type] = rel_df
-                    self.nodes[node_type] = (self.nodes[node_type]
-                                                 .drop(columns='index'))
+                    self.rels[rel_type] = rel_df[['src', 'tgt']]
+                    self.nodes[node_type] = self.nodes[node_type]
+
+            if 'index' in self.nodes[node_type]:
+                self.nodes[node_type] = (self.nodes[node_type]
+                                             .drop(columns='index'))
 
         return self
 
