@@ -47,9 +47,10 @@ def build_dgl_dataset(nodes: Dict[str, pd.DataFrame],
     for canonical_etype, rel_arr in relations.items():
         rel_arr = (relations[canonical_etype][['src', 'tgt']].drop_duplicates()
                                                              .to_numpy())
-        src_tensor = torch.from_numpy(rel_arr[:, 0]).int()
-        tgt_tensor = torch.from_numpy(rel_arr[:, 1]).int()
-        graph_data[canonical_etype] = (src_tensor, tgt_tensor)
+        if rel_arr.size:
+            src_tensor = torch.from_numpy(rel_arr[:, 0]).int()
+            tgt_tensor = torch.from_numpy(rel_arr[:, 1]).int()
+            graph_data[canonical_etype] = (src_tensor, tgt_tensor)
 
     dgl_graph = dgl.heterograph(graph_data)
 
@@ -72,14 +73,14 @@ def build_dgl_dataset(nodes: Dict[str, pd.DataFrame],
     # Add node features to the Reply nodes
     if 'reply' in nodes.keys() and 'reply' in dgl_graph.ntypes:
         cols = ['num_retweets', 'num_replies', 'num_quote_tweets']
-        tweet_feats = torch.from_numpy(nodes['reply'][cols].to_numpy())
+        reply_feats = torch.from_numpy(nodes['reply'][cols].to_numpy())
         if ('text_emb' in nodes['reply'].columns and
                 'lang_emb' in nodes['reply'].columns):
-            tweet_embs = emb_to_tensor(nodes['reply'], 'text_emb')
+            reply_embs = emb_to_tensor(nodes['reply'], 'text_emb')
             lang_embs = emb_to_tensor(nodes['reply'], 'lang_emb')
-            tensors = (tweet_embs, lang_embs, tweet_feats)
+            tensors = (reply_embs, lang_embs, reply_feats)
         else:
-            tensors = (tweet_feats,)
+            tensors = (reply_feats,)
         dgl_graph.nodes['reply'].data['feat'] = torch.cat(tensors, dim=1)
 
     # Add node features to the User nodes
